/**
 * An event that fires its listeners in a synchronous fashion.
 */
export class Event {
    /**
     * Create a new event.
     *
     * @param parent
     *   the parent that will be passed to listener as their `this`
     */
    constructor(parent) {
        this.parent = parent;
        const subscribable = (listener) => this.subscribe(listener);
        subscribable.subscribe = subscribable;
        subscribable.unsubscribe = (listener) => this.unsubscribe(listener);
        subscribable.once = () => this.once();
        this.subscribable = subscribable;
    }
    /**
     * Emit this event. This will invoke all of the listeners with the passed
     * arguments.
     *
     * @param args
     */
    emit(...args) {
        if (Array.isArray(this.listeners)) {
            /*
             * Array is present, iterate over array and invoke all of the
             * listeners.
             */
            for (const listener of this.listeners) {
                listener.apply(this.parent, args);
            }
        }
        else if (this.listeners) {
            /*
             * Single listener is present, simply invoke the listener.
             */
            this.listeners.apply(this.parent, args);
        }
    }
    /**
     * Subscribe to this event using the given listener. The listener will
     * be invoked any time the event is emitted. The returned handle can be
     * used to unsubscribe.
     *
     * @param listener
     */
    subscribe(listener) {
        if (Array.isArray(this.listeners)) {
            // Listeners is already an array, create a copy with the new listener appended
            this.listeners = [...this.listeners, listener];
        }
        else if (this.listeners) {
            this.listeners = [this.listeners, listener];
        }
        else {
            this.listeners = listener;
        }
        const self = this;
        return {
            unsubscribe() {
                return self.unsubscribe(listener);
            }
        };
    }
    /**
     * Unsubscribe a listener from this handler. The specified listener will
     * no longer be invoked when the event is emitted.
     *
     * @param listener
     */
    unsubscribe(listener) {
        if (Array.isArray(this.listeners)) {
            /*
             * Array has been allocated, find the index of the listener and
             * then remove it from the array.
             */
            const idx = this.listeners.indexOf(listener);
            if (idx < 0)
                return false;
            // Copy-on-write for deletions
            const listeners = [...this.listeners];
            listeners.splice(idx);
            this.listeners = listener;
            /*
             * If the array is empty, remove it. Otherwise at this point the
             * array has already been allocated so keep the array in case a
             * subscription happens again.
             */
            if (this.listeners.length === 0) {
                this.listeners = undefined;
            }
            return true;
        }
        else if (this.listeners === listener) {
            /*
             * Single listener is present and its the current match. Reset
             * listeners property.
             */
            this.listeners = undefined;
            return true;
        }
        // Listener is not active
        return false;
    }
    /**
     * Get a promise that will resolve the first time this event is fired
     * after this call.
     */
    once() {
        return new Promise((resolve, reject) => {
            const listener = (...args) => {
                this.unsubscribe(listener);
                resolve(args);
            };
            this.subscribe(listener);
        });
    }
}
//# sourceMappingURL=sync.js.map