import { SubscriptionHandle } from './handle';
/**
 * Listener type. Defines the function signature that a listener is expected
 * to have.
 */
export declare type Listener<This, Args extends any[]> = (this: This, ...args: Args) => void;
/**
 * Functions used to subscribe and unsubscribe to an event.
 */
export interface Subscribable<This, Args extends any[]> {
    /**
     * Subscribe to the event, will invoke the given function when the event
     * is emitted.
     */
    (listener: Listener<This, Args>): SubscriptionHandle;
    /**
     * Subscribe to the event, will invoke the given function when the event
     * is emitted.
     */
    subscribe(listener: Listener<This, Args>): SubscriptionHandle;
    /**
     * Unsubscribe a previously subscribed listener.
     *
     * @param listener
     */
    unsubscribe(listener: Listener<This, Args>): boolean;
    /**
     * Subscribe to an event but only trigger the listener once.
     *
     * @param listener
     */
    once(): Promise<Args>;
}
/**
 * An event that fires its listeners in a synchronous fashion.
 */
export declare class Event<Parent, Args extends any[] = []> {
    /**
     * Public Subscribable that can safely be shared with consumers that should
     * be able to listen for events.
     */
    readonly subscribable: Subscribable<Parent, Args>;
    /**
     * Parent of this handler, used to apply the correct this to event listeners.
     */
    private readonly parent;
    /**
     * Listener(s) that have been attached to this event handler.
     */
    private listeners?;
    /**
     * Create a new event.
     *
     * @param parent
     *   the parent that will be passed to listener as their `this`
     */
    constructor(parent: Parent);
    /**
     * Emit this event. This will invoke all of the listeners with the passed
     * arguments.
     *
     * @param args
     */
    emit(...args: Args): void;
    /**
     * Subscribe to this event using the given listener. The listener will
     * be invoked any time the event is emitted. The returned handle can be
     * used to unsubscribe.
     *
     * @param listener
     */
    subscribe(listener: Listener<Parent, Args>): SubscriptionHandle;
    /**
     * Unsubscribe a listener from this handler. The specified listener will
     * no longer be invoked when the event is emitted.
     *
     * @param listener
     */
    unsubscribe(listener: Listener<Parent, Args>): boolean;
    /**
     * Get a promise that will resolve the first time this event is fired
     * after this call.
     */
    once(): Promise<Args>;
}
//# sourceMappingURL=sync.d.ts.map